"""
    sphinxit.core.connector
    ~~~~~~~~~~~~~~~~~~~~~~~

    Implements Sphinxit <-> searchd interaction.

    :copyright: (c) 2013 by Roman Semirook.
    :license: BSD, see LICENSE for more details.
"""

from __future__ import unicode_literals

import threading
from collections import deque

from .mixins import ConfigMixin
from .exceptions import ImproperlyConfigured, SphinxQLDriverException

oursql = False
mysqldb = False

try:
    import oursql as mysql_client
    oursql = True
except ImportError:
    try:
        import MySQLdb as mysql_client
        import MySQLdb.cursors
        mysqldb = True
    except ImportError:
        pass


class SphinxConnector(ConfigMixin):

    def __init__(self, config):
        connection_options = {
            'host': '127.0.0.1',
            'port': 9306,
        }
        connection_options.update(config.SEARCHD_CONNECTION)

        self.config = config
        self.connection_options = connection_options

        self.__connections_pool = deque([])
        self.__local = threading.local()
        self.__conn_lock = threading.Lock()

    def __del__(self):
        self.close_connections()

    def close_connections(self):
        with self.__conn_lock:
            if hasattr(self.__local, 'conn'):
                del self.__local.conn
            for connection in self.__connections_pool:
                del connection

    def get_connection(self):
        if not oursql and not mysqldb:
            raise ImproperlyConfigured(
                'Oursql or MySQLdb library has to be installed to work with searchd'
            )
        if not self.__connections_pool:
            for i in range(getattr(self.config, 'POOL_SIZE', 10)):
                if oursql:
                    self.__connections_pool.append(mysql_client.connect(
                        **self.connection_options
                    ))
                if mysqldb:
                    self.__connections_pool.append(mysql_client.connect(
                        cursorclass=mysql_client.cursors.DictCursor,
                        use_unicode=self.connection_options.pop('use_unicode', True),
                        charset=self.connection_options.pop('charset', 'utf8'),
                        **self.connection_options
                    ))

        with self.__conn_lock:
            self.__local.conn = self.__connections_pool.pop()

        return self.__local.conn

    def get_cursor(self, connection):
        if oursql:
            curs = connection.cursor(mysql_client.DictCursor)
        if mysqldb:
            curs = connection.cursor()

        return curs

    def _get_cursor_exec(self, curs):
        if oursql:
            execute_query = lambda sxql_query: curs.execute(sxql_query, plain_query=True)
        if mysqldb:
            execute_query = lambda sxql_query: curs.execute(sxql_query)

        return execute_query

    def _normalize_meta(self, raw_result):
        return dict([(x['Variable_name'], x['Value']) for x in raw_result])

    def _execute_batch(self, cursor, sxql_batch):
        total_results = {}
        extra = []
        if getattr(self.config, 'WITH_META', False):
            extra.append(('SHOW META', 'meta'))
        if getattr(self.config, 'WITH_STATUS', False):
            extra.append(('SHOW STATUS', 'status'))

        cursor_exec = self._get_cursor_exec(cursor)

        for sub_ql_pair in sxql_batch:
            subresult = {}
            sub_ql, sub_alias = sub_ql_pair
            cursor_exec(sub_ql)
            subresult['items'] = [r for r in cursor]

            for sub_meta_pair in extra:
                meta_ql, meta_alias = sub_meta_pair
                cursor_exec(meta_ql)
                subresult[meta_alias] = self._normalize_meta(cursor)

            total_results[sub_alias] = subresult

        return total_results

    def _execute_query(self, cursor, sxql_query):
        cursor_exec = self._get_cursor_exec(cursor)
        cursor_exec(sxql_query)

        return cursor.fetchall()

    def execute(self, sxql_query):
        connection = self.get_connection()
        cursor = self.get_cursor(connection)
        total_results = {}
        try:
            if isinstance(sxql_query, (tuple, list)):
                total_results = self._execute_batch(cursor, sxql_query)
            else:
                total_results = self._execute_query(cursor, sxql_query)
        except Exception as e:
            if oursql and type(e).__name__ == 'ProgrammingError':
                errno, msg, extra = e
                if errno is not None:
                    raise SphinxQLDriverException(msg)
            else:
                raise SphinxQLDriverException(e)
        finally:
            cursor.close()
            self.__connections_pool.appendleft(connection)

        return total_results



"""
    sphinxit.core.constants
    ~~~~~~~~~~~~~~~~~~~~~~~

    Defines some Sphinx-specific constants.

    :copyright: (c) 2013 by Roman Semirook.
    :license: BSD, see LICENSE for more details.
"""

from collections import namedtuple


RESERVED_KEYWORDS = (
    'AND',
    'AS',
    'ASC',
    'AVG',
    'BEGIN',
    'BETWEEN',
    'BY',
    'CALL',
    'COLLATION',
    'COMMIT',
    'COUNT',
    'DELETE',
    'DESC',
    'DESCRIBE',
    'DISTINCT',
    'FALSE',
    'FROM',
    'GLOBAL',
    'GROUP',
    'IN',
    'INSERT',
    'INTO',
    'LIMIT',
    'MATCH',
    'MAX',
    'META',
    'MIN',
    'NOT',
    'NULL',
    'OPTION',
    'OR',
    'ORDER',
    'REPLACE',
    'ROLLBACK',
    'SELECT',
    'SET',
    'SHOW',
    'START',
    'STATUS',
    'SUM',
    'TABLES',
    'TRANSACTION',
    'TRUE',
    'UPDATE',
    'VALUES',
    'VARIABLES',
    'WARNINGS',
    'WEIGHT',
    'WHERE',
    'WITHIN'
)


ESCAPED_CHARS = namedtuple('EscapedChars', ['single_escape', 'double_escape'])(
    single_escape=("'", '+', '[', ']', '=', '*'),
    double_escape=('@', '!', '^', '(', ')', '~', '-', '|', '/', '<<', '$', '"')
)

NODES_ORDER = namedtuple('NodesOrder', ['select', 'update'])(
    select=(
        'SelectFrom',
        'Where',
        'GroupBy',
        'OrderBy',
        'WithinGroupOrderBy',
        'Limit',
        'Options'
    ),
    update=(
        'UpdateSet',
        'Where',
        'Options'
    )
)



"""
    sphinxit.core.nodes
    ~~~~~~~~~~~~~~~~~~~

    Implements atomic nodes and containers of lexemes.

    :copyright: (c) 2013 by Roman Semirook.
    :license: BSD, see LICENSE for more details.
"""

from __future__ import unicode_literals

from collections import deque

from sphinxit.core.convertors import (
    FilterCtx,
    ORFilterCtx,
    MatchQueryCtx,
    AliasFieldCtx,
    FieldCtx,
    OrderCtx,
    LimitCtx,
    OptionsCtx,
    UpdateSetCtx,
    SnippetsOptionsCtx
)
from sphinxit.core.helpers import (
    sparse_free_sequence,
    string_from_string,
)
from sphinxit.core.exceptions import SphinxQLSyntaxException
from sphinxit.core.mixins import ConfigMixin


class SelectFromContainer(ConfigMixin):
    _joiner = ', '
    _template = 'SELECT {fields} FROM {indexes}'

    def __init__(self, indexes=None):
        super(SelectFromContainer, self).__init__()
        self.indexes = indexes
        self.fields = []
        self.or_fields = []

    def add_alias(self, field, alias):
        with AliasFieldCtx(field, alias).with_config(self.config) as lex:
            if lex and lex not in self.fields:
                self.fields.append(lex)

    def add_field(self, field):
        with FieldCtx(field).with_config(self.config) as lex:
            if lex and lex not in self.fields:
                self.fields.append(lex)

    def add_or(self, or_instance):
        assert isinstance(or_instance, OR), type(or_instance)
        with AliasFieldCtx(
            or_instance.with_config(self.config).lex(), 'cnd'
        ).with_config(self.config) as lex:
            if lex and lex not in self.or_fields:
                self.or_fields.append(lex)

    def add_aggregation(self, agg_instance):
        assert isinstance(agg_instance, AggregateObject), type(agg_instance)
        agg_lex = agg_instance.with_config(self.config).lex()
        if agg_lex and agg_lex not in self.fields:
            self.fields.append(agg_lex)

    def add_raw_attr(self, raw_attr_instance):
        assert isinstance(raw_attr_instance, RawAttr), type(raw_attr_instance)
        raw_lex = raw_attr_instance.with_config(self.config).lex()
        if raw_lex and raw_lex not in self.fields:
            self.fields.append(raw_lex)

    def has_or_fields(self):
        return bool(self.or_fields)

    def lex(self):
        if self.indexes is None:
            raise SphinxQLSyntaxException('No indexes defined to search with')

        lex = self._template.format(
            fields=self._joiner.join((self.fields or ['*']) + self.or_fields),
            indexes=self._joiner.join(self.indexes),
        )
        return lex


class FiltersContainer(ConfigMixin):
    _joiner = ' AND '
    _match_template = "MATCH('{query}')"
    _where_template = 'WHERE {conditions}'

    def __init__(self):
        super(FiltersContainer, self).__init__()
        self.query = []
        self.conditions = []

    def __bool__(self):
        return bool(self.conditions or self.query)

    def add_query(self, query):
        with MatchQueryCtx(query).with_config(self.config) as lex:
            if lex:
                self.query.append(lex)

    def add_raw_query(self, query):
        self.query.append(query)

    def add_condition(self, field, value):
        with FilterCtx(field, value).with_config(self.config) as lex:
            if lex and lex not in self.conditions:
                self.conditions.append(lex)

    def add_conditions(self, **kwargs):
        for field, value in kwargs.items():
            self.add_condition(field, value)

    def lex(self):
        query_lex = ''
        cond_lex = ''
        if self.query:
            query_lex = self._match_template.format(
                query=' '.join(self.query)
            )
        if self.conditions:
            cond_lex = self._joiner.join(self.conditions)

        return self._where_template.format(
            conditions=self._joiner.join(sparse_free_sequence([query_lex, cond_lex]))
        )


class GroupByNode(ConfigMixin):
    _template = 'GROUP BY {field}'

    def __init__(self):
        super(GroupByNode, self).__init__()
        self.field = None

    def __bool__(self):
        return bool(self.field)

    def by_field(self, field):
        if not self:
            with FieldCtx(field).with_config(self.config) as lex:
                if lex:
                    self.field = field

    def lex(self):
        if self:
            return self._template.format(field=self.field)
        return ''


class OrderByContainer(ConfigMixin):
    _joiner = ', '
    _template = 'ORDER BY {orderings}'

    def __init__(self):
        super(OrderByContainer, self).__init__()
        self.orderings = set()

    def __bool__(self):
        return bool(self.orderings)

    def by_field(self, field, direction='ASC'):
        with OrderCtx(field, direction).with_config(self.config) as lex:
            if lex and lex not in self.orderings:
                self.orderings.add(lex)

    def lex(self):
        if self:
            return self._template.format(
                orderings=self._joiner.join(self.orderings)
            )
        return ''


class WithinGroupOrderByNode(ConfigMixin):
    _template = 'WITHIN GROUP ORDER BY {field}'

    def __init__(self):
        super(WithinGroupOrderByNode, self).__init__()
        self.field = None

    def __bool__(self):
        return bool(self.field)

    def by_field(self, field, direction='ASC'):
        if not self:
            with OrderCtx(field, direction).with_config(self.config) as lex:
                if lex:
                    self.field = lex

    def lex(self):
        if self:
            return self._template.format(field=self.field)
        return ''


class LimitNode(ConfigMixin):
    _template = 'LIMIT {offset},{limit}'

    def __init__(self):
        super(LimitNode, self).__init__()
        self.offset = None
        self.limit = None

    def __bool__(self):
        return self.offset is not None and self.limit is not None

    def set_range(self, offset, limit):
        if not self:
            with LimitCtx(offset, limit).with_config(self.config) as pair:
                self.offset, self.limit = pair

    def lex(self):
        if self:
            return self._template.format(
                offset=self.offset,
                limit=self.limit
            )
        return ''


class OptionsContainer(ConfigMixin):
    _joiner = ', '
    _template = 'OPTION {options}'

    def __init__(self):
        super(OptionsContainer, self).__init__()
        self.options = deque([])

    def __bool__(self):
        return bool(self.options)

    def set_options(self, **kwargs):
        for option, params in kwargs.items():
            with OptionsCtx(option, params).with_config(self.config) as lex:
                if lex:
                    self.options.append(lex)

    def add_ranker(self, ranker):
        with OptionsCtx(
            'ranker', ranker
        ).with_config(self.config) as lex:
            if lex:
                self.options.append(lex)

    def add_max_matches(self, max_matches):
        with OptionsCtx(
            'max_matches', max_matches
        ).with_config(self.config) as lex:
            if lex:
                self.options.append(lex)

    def add_cutoff(self, cutoff):
        with OptionsCtx(
            'cutoff', cutoff
        ).with_config(self.config) as lex:
            if lex:
                self.options.append(lex)

    def add_max_query_time(self, max_query_time):
        with OptionsCtx(
            'max_query_time', max_query_time
        ).with_config(self.config) as lex:
            if lex:
                self.options.append(lex)

    def add_retry_count(self, retry_count):
        with OptionsCtx(
            'retry_count', retry_count
        ).with_config(self.config) as lex:
            if lex:
                self.options.append(lex)

    def add_retry_delay(self, retry_delay):
        with OptionsCtx(
            'retry_delay', retry_delay
        ).with_config(self.config) as lex:
            if lex:
                self.options.append(lex)

    def add_field_weights(self, **kwargs):
        with OptionsCtx(
            'field_weights', kwargs
        ).with_config(self.config) as lex:
            if lex:
                self.options.append(lex)

    def add_index_weights(self, **kwargs):
        with OptionsCtx(
            'index_weights', kwargs
        ).with_config(self.config) as lex:
            if lex:
                self.options.append(lex)

    def add_reverse_scan(self, is_reverse=True):
        with OptionsCtx(
            'reverse_scan', is_reverse
        ).with_config(self.config) as lex:
            if lex:
                self.options.append(lex)

    def add_comment(self, comment):
        with OptionsCtx(
            'comment', comment
        ).with_config(self.config) as lex:
            if lex:
                self.options.append(lex)

    def lex(self):
        if self:
            return self._template.format(
                options=self._joiner.join(self.options)
            )
        return ''


class UpdateSetNode(ConfigMixin):
    _template = 'UPDATE {indexes} SET {values}'
    _joiner = ', '

    def __init__(self, indexes=None):
        super(UpdateSetNode, self).__init__()
        self.indexes = indexes
        self.set_values = set()

    def update(self, field, value):
        with UpdateSetCtx(field, value).with_config(self.config) as lex:
            self.set_values.add(lex)

    def lex(self):
        if self.set_values:
            return self._template.format(
                indexes=self._joiner.join(self.indexes),
                values=self._joiner.join(self.set_values)
            )
        return ''


class OR(ConfigMixin):
    _wrapper = '(%s)'
    _joiner = ' OR '

    def __init__(self, **kwargs):
        super(OR, self).__init__()
        self.raw_attrs = kwargs
        self.children = []
        self.joiner = None

    def __join(self, or_inst, joiner):
        or_ = OR()
        or_.children.extend([self, or_inst])
        or_.joiner = joiner
        return or_

    def __or__(self, other):
        return self.__join(other, joiner=' OR ')

    def __and__(self, other):
        return self.__join(other, joiner=' AND ')

    def lex(self):
        flat_conditions = []
        flat_joiners = []

        def expand_tree(self):
            if self.joiner is not None:
                flat_joiners.append(self.joiner)
            if self.raw_attrs:
                cleaned_conditions = []
                for k_attr, v_attr in self.raw_attrs.items():
                    with ORFilterCtx(k_attr, v_attr).with_config(self.config) as lex:
                        if lex:
                            cleaned_conditions.append(lex)
                if cleaned_conditions:
                    flat_conditions.append(
                        self._wrapper % self._joiner.join(cleaned_conditions)
                    )
            for c in self.children:
                expand_tree(
                    c.with_config(self.config)
                    if not c.has_config()
                    else c
                )
            flat_joiners.reverse()
            return flat_conditions, flat_joiners

        flat_conditions, flat_joiners = expand_tree(self)
        joiners = iter(flat_joiners)
        lex = reduce(
            lambda x, y: next(joiners).join([x, y]),
            flat_conditions
        )
        return lex


class AggregateObject(ConfigMixin):
    _agg_template = None
    _alias_template = None

    def __init__(self, field=None, alias=None):
        super(AggregateObject, self).__init__()
        if self._agg_template is None or self._alias_template is None:
            raise NotImplementedError(
                'Im am AggregateObject base class, '
                'inherit me and implement what you need'
            )

        self.raw_attrs = (field, alias or self._alias_template.format(field=field))

    def lex(self):
        raw_field, raw_alias = self.raw_attrs
        with AliasFieldCtx(
            raw_field, raw_alias
        ).called_by(
            self.__class__
        ).with_config(
            self.config
        ) as lex:
            return '' if not lex else self._agg_template.format(
                field=raw_field,
                alias=raw_alias,
            )


class RawAttr(ConfigMixin):
    _template = '{field} AS {alias}'

    def __init__(self, field, alias):
        self.field = field
        self.alias = alias

    def lex(self):
        with AliasFieldCtx(
            self.field, self.alias
        ).called_by(
            self.__class__
        ).with_config(
            self.config
        ) as lex:
            return '' if not lex else self._template.format(
                field=self.field,
                alias=self.alias
            )
        return


class Avg(AggregateObject):
    _agg_template = 'AVG({field}) AS {alias}'
    _alias_template = '{field}_avg'


class Min(AggregateObject):
    _agg_template = 'MIN({field}) AS {alias}'
    _alias_template = '{field}_min'


class Max(AggregateObject):
    _agg_template = 'MAX({field}) AS {alias}'
    _alias_template = '{field}_max'


class Sum(AggregateObject):
    _agg_template = 'SUM({field}) AS {alias}'
    _alias_template = '{field}_sum'


class Count(AggregateObject):
    _agg_template = 'COUNT(DISTINCT {field}) AS {alias}'
    _star_agg_template = 'COUNT({field}) AS {alias}'
    _alias_template = '{field}_count'

    def __init__(self, field='*', alias='num'):
        super(Count, self).__init__(field, alias)

    def lex(self):
        raw_field, raw_alias = self.raw_attrs
        with AliasFieldCtx(
            raw_field, raw_alias
        ).called_by(
            self.__class__
        ).with_config(
            self.config
        ) as lex:
            if raw_field == '*':
                template = self._star_agg_template
            else:
                template = self._agg_template
                if raw_alias == 'num':
                    raw_alias = self._alias_template.format(field=raw_field)

            return '' if not lex else template.format(field=raw_field, alias=raw_alias)


class SnippetsOptionsContainer(ConfigMixin):
    _joiner = ', '

    def __init__(self):
        super(SnippetsOptionsContainer, self).__init__()
        self.options = []

    def __bool__(self):
        return bool(self.options)

    def set_options(self, **kwargs):
        for option, params in kwargs.items():
            with SnippetsOptionsCtx(option, params).with_config(self.config) as lex:
                if lex and lex not in self.options:
                    self.options.append(lex)

    def lex(self):
        if self:
            return self._joiner.join(self.options)
        return ''


class SnippetsQueryNode(ConfigMixin):
    _joiner = ', '
    _template = "{data}, '{index}', '{query}'"

    def __init__(self, index=None):
        super(SnippetsQueryNode, self).__init__()
        self.index = index
        self.data = []
        self.query = []

    def __bool__(self):
        return all((self.index, self.data, self.query))

    def add_data(self, *data):
        data = sparse_free_sequence(data)
        for value in data:
            value = string_from_string(value, self.is_strict)
            if value and value not in self.data:
                self.data.append(value)
        return self

    def add_query(self, query):
        with MatchQueryCtx(query).with_config(self.config) as lex:
            if lex and lex not in self.query:
                self.query.append(lex)
        return self

    def lex(self):
        if not self:
            return ''

        if len(self.data) == 1:
            data_wrapper = "{data}"
        else:
            data_wrapper = "({data})"

        data = data_wrapper.format(
            data=self._joiner.join(["'%s'" % d for d in self.data])
        )

        return self._template.format(
            data=data,
            index=self.index,
            query=' '.join(self.query)
        )



from __future__ import unicode_literals

import random
from sphinxit.core.constants import RESERVED_KEYWORDS

try:
    import unittest2 as unittest
except ImportError:
    import unittest

from sphinxit.core.nodes import (
    OR,
    Avg,
    Min,
    Max,
    Sum,
    Count,
    SnippetsQueryNode
)


class DebugConfig(object):
    DEBUG = True


class ProductionConfig(object):
    DEBUG = False


ORSoft = lambda **kwargs: OR(**kwargs).with_config(ProductionConfig)
ORStrict = lambda **kwargs: OR(**kwargs).with_config(DebugConfig)

class TestORNode(unittest.TestCase):

    def test_self(self):
        or_ = ORSoft(age__gte=14)
        self.assertEqual(or_.lex(), '(age>=14)')

    def test_double_self(self):
        or_ = ORSoft(age__gte=14, age__eq=3)
        possible_results = ('(age>=14 OR age=3)', '(age=3 OR age>=14)')
        self.assertIn(or_.lex(), possible_results)

    def test_and_join(self):
        or_ = (
            OR(age__gte=14, age__eq=3)
            & OR(rank__eq=1)
            & OR(rank__eq=2)
            | OR(count__eq=1, count__lte=100, count__between=(1, 10))
        ).with_config(ProductionConfig)
        possible_results = (
            u'(age=3 OR age>=14) AND (rank=1) AND (rank=2) OR (count=1 OR count<=100)',
            u'(age=3 OR age>=14) AND (rank=1) AND (rank=2) OR (count<=100 OR count=1)',
            u'(age>=14 OR age=3) AND (rank=1) AND (rank=2) OR (count=1 OR count<=100)',
            u'(age>=14 OR age=3) AND (rank=1) AND (rank=2) OR (count<=100 OR count=1)',
        )
        self.assertIn(or_.lex(), possible_results)

    def test_or_join(self):
        or_ = (
            OR(age__gte=14, age__eq=3) |
            OR(count__eq=1, count__gte=100, count__between=(1, 10))
        ).with_config(ProductionConfig)
        possible_results = (
            '(age=3 OR age>=14) OR (count=1 OR count>=100)',
            '(age=3 OR age>=14) OR (count>=100 OR count=1)',
            '(age>=14 OR age=3) OR (count=1 OR count>=100)',
            '(age>=14 OR age=3) OR (count>=100 OR count=1)',
        )
        self.assertIn(or_.lex(), possible_results)


AvgSoft = lambda *args, **kwargs: Avg(*args, **kwargs).with_config(ProductionConfig).lex()
AvgStrict = lambda *args, **kwargs: Avg(*args, **kwargs).with_config(DebugConfig).lex()

class TestAvg(unittest.TestCase):

    def test_avg_empty(self):
        self.assertFalse(AvgSoft())

    def test_avg_partly_empty(self):
        self.assertEqual(AvgSoft(field='name'), 'AVG(name) AS name_avg')

    def test_avg_partly_empty_with_alias(self):
        self.assertFalse(AvgSoft(alias='name'))

    def test_avg_complete(self):
        self.assertEqual(
            AvgSoft(field='name', alias='another_name'),
            'AVG(name) AS another_name'
        )


MinSoft = lambda *args, **kwargs: Min(*args, **kwargs).with_config(ProductionConfig).lex()
MinStrict = lambda *args, **kwargs: Min(*args, **kwargs).with_config(DebugConfig).lex()

class TestMin(unittest.TestCase):

    def test_min_empty(self):
        self.assertFalse(MinSoft())

    def test_min_partly_empty(self):
        self.assertEqual(MinSoft(field='name'), 'MIN(name) AS name_min')

    def test_min_partly_empty_with_alias(self):
        self.assertFalse(MinSoft(alias='name'))

    def test_min_complete(self):
        self.assertEqual(
            MinSoft(field='name', alias='another_name'),
            'MIN(name) AS another_name'
        )


MaxSoft = lambda *args, **kwargs: Max(*args, **kwargs).with_config(ProductionConfig).lex()
MaxStrict = lambda *args, **kwargs: Max(*args, **kwargs).with_config(DebugConfig).lex()

class TestMax(unittest.TestCase):

    def test_max_empty(self):
        self.assertFalse(MaxSoft())

    def test_max_partly_empty(self):
        self.assertEqual(MaxSoft(field='name'), 'MAX(name) AS name_max')

    def test_max_partly_empty_with_alias(self):
        self.assertFalse(MaxSoft(alias='name'))

    def test_max_complete(self):
        self.assertEqual(
            MaxSoft(field='name', alias='another_name'),
            'MAX(name) AS another_name'
        )


SumSoft = lambda *args, **kwargs: Sum(*args, **kwargs).with_config(ProductionConfig).lex()
SumStrict = lambda *args, **kwargs: Sum(*args, **kwargs).with_config(DebugConfig).lex()

class TestSum(unittest.TestCase):

    def test_sum_empty(self):
        self.assertFalse(SumSoft())

    def test_sum_partly_empty(self):
        self.assertEqual(SumSoft(field='name'), 'SUM(name) AS name_sum')

    def test_sum_partly_empty_with_alias(self):
        self.assertFalse(SumSoft(alias='name'))

    def test_sum_complete(self):
        self.assertEqual(
            SumSoft(field='name', alias='another_name'),
            'SUM(name) AS another_name'
        )


CountSoft = lambda *args, **kwargs: Count(*args, **kwargs).with_config(ProductionConfig)
CountStrict = lambda *args, **kwargs: Count(*args, **kwargs).with_config(DebugConfig)

class TestCount(unittest.TestCase):

    def test_count_empty(self):
        self.assertEqual(CountSoft().lex(), 'COUNT(*) AS num')

    def test_count_with_alias(self):
        self.assertEqual(CountSoft(alias='counter').lex(), 'COUNT(*) AS counter')

    def test_count_with_field(self):
        self.assertEqual(
            CountSoft(field='age').lex(),
            'COUNT(DISTINCT age) AS age_count'
        )

    def test_count_complete(self):
        self.assertEqual(
            CountSoft(field='age', alias='age_num').lex(),
            'COUNT(DISTINCT age) AS age_num'
        )

    def test_count_with_invalid_field(self):
        self.assertFalse(
            CountSoft(field=random.choice(RESERVED_KEYWORDS), alias='alias').lex(),
        )

    def test_count_with_invalid_alias(self):
        self.assertFalse(
            CountSoft(field='age', alias=random.choice(RESERVED_KEYWORDS)).lex(),
        )


SnippetsQueryNodeSoft = lambda *args, **kwargs: SnippetsQueryNode(*args, **kwargs).with_config(ProductionConfig)
SnippetsQueryNodeStrict = lambda *args, **kwargs: SnippetsQueryNode(*args, **kwargs).with_config(DebugConfig)

class TestSnippetsQueryNode(unittest.TestCase):

    def test_valid_attrs(self):
        node = (
            SnippetsQueryNodeSoft(index='index_name')
            .add_data('me amore sphinx')
            .add_query('me amore')
        )
        self.assertEqual(
            node.lex(),
            "'me amore sphinx', 'index_name', 'me amore'"
        )

    def test_valid_extended_data_attrs(self):
        node = (
            SnippetsQueryNodeSoft(index='index_name')
            .add_data('me amore sphinx', 'me amore python')
            .add_query('me amore')
        )
        self.assertEqual(
            node.lex(),
            "('me amore sphinx', 'me amore python'), 'index_name', 'me amore'"
        )

    def test_valid_extended_query_attrs(self):
        node = (
            SnippetsQueryNodeSoft(index='index_name')
            .add_data('me amore', 'me amore python')
            .add_query('me amore')
            .add_query('python')
        )
        self.assertEqual(
            node.lex(),
            "('me amore', 'me amore python'), 'index_name', 'me amore python'"
        )

    def test_valid_extended_attrs(self):
        node = (
            SnippetsQueryNodeSoft(index='index_name')
            .add_data('me amore', 'me amore python')
            .add_data('me amore python')
            .add_query('me amore')
            .add_query('me amore')
            .add_query('python')
        )
        self.assertEqual(
            node.lex(),
            "('me amore', 'me amore python'), 'index_name', 'me amore python'"
        )

    def test_incomplete_attrs(self):
        self.assertEqual(SnippetsQueryNodeSoft().lex(), '')
        self.assertEqual(SnippetsQueryNodeSoft(index='index_name').lex(), '')
        self.assertEqual(
            SnippetsQueryNodeSoft(index='index_name').add_data('me amore').lex(), ''
        )
        self.assertEqual(
            SnippetsQueryNodeSoft(index='index_name').add_query('python').lex(), ''
        )

    def test_partly_invalid_data(self):
        node = (
            SnippetsQueryNodeSoft(index='index_name')
            .add_data('me amore sphinx', None, 42)
            .add_query('amore')
        )
        self.assertEqual(node.lex(), "'me amore sphinx', 'index_name', 'amore'")

    def test_invalid_data(self):
        node = (
            SnippetsQueryNodeSoft(index='index_name')
            .add_data(None, 42, object, [''], '   ')
            .add_query('   ')
            .add_query('query')
        )
        self.assertEqual(node.lex(), '')



from __future__ import unicode_literals

import random
from datetime import datetime, date

try:
    import unittest2 as unittest
except ImportError:
    import unittest

from sphinxit.core.constants import RESERVED_KEYWORDS, ESCAPED_CHARS
from sphinxit.core.exceptions import SphinxQLSyntaxException
from sphinxit.core.helpers import (
    list_of_integers_only,
    int_from_digit,
    sparse_free_sequence,
    unix_timestamp
)
from sphinxit.core.convertors import (
    FilterCtx,
    MatchQueryCtx,
    AliasFieldCtx,
    LimitCtx,
    OrderCtx,
    FieldCtx,
    OptionsCtx,
    UpdateSetCtx,
)


class DebugConfig(object):
    DEBUG = True


class ProductionConfig(object):
    DEBUG = False


class TestListOfIntegersOnlyConverter(unittest.TestCase):

    def test_clean_sequence(self):
        sequence = [1, 2, 3, 4, 5]
        self.assertListEqual(list_of_integers_only(sequence), sequence)

    def test_digit_str_sequence(self):
        sequence = [1, '2', 3, 4, '5']
        self.assertListEqual(list_of_integers_only(sequence), [1, 2, 3, 4, 5])

    def test_str_sequence(self):
        sequence = [1, 'boom', 3, 4, '5']
        self.assertListEqual(list_of_integers_only(sequence), [1, 3, 4, 5])

    def test_str_sequence_strict(self):
        sequence = [1, 'boom', 3, 4, '5']
        self.assertRaises(
            SphinxQLSyntaxException,
            lambda: list_of_integers_only(sequence, is_strict=True),
        )


class TestIntFromDigit(unittest.TestCase):

    def test_int(self):
        self.assertEqual(int_from_digit(42), 42)

    def test_digit_str(self):
        self.assertEqual(int_from_digit('42'), 42)

    def test_str(self):
        self.assertIsNone(int_from_digit('boom'))

    def test_str_strict(self):
        self.assertRaises(
            SphinxQLSyntaxException,
            lambda: int_from_digit('boom', is_strict=True),
        )


class TestSparseFreeSequence(unittest.TestCase):

    def test_sparse(self):
        dirty_list = ['1', 2, None, '4', '  ', 6, {'key': 'value'}, [], ()]
        self.assertEqual(
            sparse_free_sequence(dirty_list),
            ['1', 2, '4', 6, {'key': 'value'}]
        )


FilterCtxStrict = lambda *args, **kwargs: FilterCtx(*args, **kwargs).with_config(DebugConfig)
FilterCtxSoft = lambda *args, **kwargs: FilterCtx(*args, **kwargs).with_config(ProductionConfig)

class TestFilterCtx(unittest.TestCase):

    def test_eq_non_strict(self):
        with FilterCtxSoft('age__eq', 25) as value:
            self.assertEqual(value, 'age=25')

        with FilterCtxSoft('age__eq', '25') as value:
            self.assertEqual(value, 'age=25')

        with FilterCtxSoft('age__eq', '25d') as value:
            self.assertIsNone(value)

    def test_eq_strict(self):
        with FilterCtxStrict('age__eq', 25) as value:
            self.assertEqual(value, 'age=25')

        with FilterCtxStrict('age__eq', '25') as value:
            self.assertEqual(value, 'age=25')

        self.assertRaises(
            SphinxQLSyntaxException,
            lambda: FilterCtxStrict('age__eq', '25d').__enter__()
        )

    def test_neq_non_strict(self):
        with FilterCtxSoft('age__neq', 25) as value:
            self.assertEqual(value, 'age!=25')

        with FilterCtxSoft('age__neq', '25') as value:
            self.assertEqual(value, 'age!=25')

        with FilterCtxSoft('age__neq', '25d') as value:
            self.assertIsNone(value)

    def test_neq_strict(self):
        with FilterCtxStrict('age__neq', 25) as value:
            self.assertEqual(value, 'age!=25')

        with FilterCtxStrict('age__neq', '25') as value:
            self.assertEqual(value, 'age!=25')

        self.assertRaises(
            SphinxQLSyntaxException,
            lambda: FilterCtxStrict('age__neq', '25d').__enter__()
        )

    def test_gt_non_strict(self):
        with FilterCtxSoft('age__gt', 25) as value:
            self.assertEqual(value, 'age>25')

        with FilterCtxSoft('age__gt', '25') as value:
            self.assertEqual(value, 'age>25')

        with FilterCtxSoft('age__gt', '25d') as value:
            self.assertIsNone(value)

    def test_gt_strict(self):
        with FilterCtxStrict('age__gt', 25) as value:
            self.assertEqual(value, 'age>25')

        with FilterCtxStrict('age__gt', '25') as value:
            self.assertEqual(value, 'age>25')

        self.assertRaises(
            SphinxQLSyntaxException,
            lambda: FilterCtxStrict('age__gt', '25d').__enter__()
        )

    def test_gte_non_strict(self):
        with FilterCtxSoft('age__gte', 25) as value:
            self.assertEqual(value, 'age>=25')

        with FilterCtxSoft('age__gte', '25') as value:
            self.assertEqual(value, 'age>=25')

        with FilterCtxSoft('age__gte', '25d') as value:
            self.assertIsNone(value)

    def test_gte_strict(self):
        with FilterCtxStrict('age__gte', 25) as value:
            self.assertEqual(value, 'age>=25')

        with FilterCtxStrict('age__gte', '25') as value:
            self.assertEqual(value, 'age>=25')

        self.assertRaises(
            SphinxQLSyntaxException,
            lambda: FilterCtxStrict('age__gte', '25d').__enter__()
        )

    def test_lt_non_strict(self):
        with FilterCtxSoft('age__lt', 25) as value:
            self.assertEqual(value, 'age<25')

        with FilterCtxSoft('age__lt', '25') as value:
            self.assertEqual(value, 'age<25')

        with FilterCtxSoft('age__lt', '25d') as value:
            self.assertEqual(value, None)

    def test_lt_strict(self):
        with FilterCtxStrict('age__lt', 25) as value:
            self.assertEqual(value, 'age<25')

        with FilterCtxStrict('age__lt', '25') as value:
            self.assertEqual(value, 'age<25')

        self.assertRaises(
            SphinxQLSyntaxException,
            lambda: FilterCtxStrict('age__lt', '25d').__enter__()
        )

    def test_lte_non_strict(self):
        with FilterCtxSoft('age__lte', 25) as value:
            self.assertEqual(value, 'age<=25')

        with FilterCtxSoft('age__lte', '25') as value:
            self.assertEqual(value, 'age<=25')

        with FilterCtxSoft('age__lte', '25d') as value:
            self.assertEqual(value, None)

    def test_lte_strict(self):
        with FilterCtxStrict('age__lte', 25) as value:
            self.assertEqual(value, 'age<=25')

        with FilterCtxStrict('age__lte', '25') as value:
            self.assertEqual(value, 'age<=25')

        self.assertRaises(
            SphinxQLSyntaxException,
            lambda: FilterCtxStrict('age__lte', '25d').__enter__()
        )

    def test_in_non_strict(self):
        with FilterCtxSoft('age__in', (1, 2, 3)) as value:
            self.assertEqual(value, 'age IN (1,2,3)')

        with FilterCtxSoft('age__in', (1, '2', 3)) as value:
            self.assertEqual(value, 'age IN (1,2,3)')

        with FilterCtxSoft('age__in', (1, '2', 'string', 3, '42d')) as value:
            self.assertEqual(value, 'age IN (1,2,3)')

        with FilterCtxSoft('age__in', []) as value:
            self.assertIsNone(value)

        with FilterCtxSoft('age__in', None) as value:
            self.assertIsNone(value)

    def test_in_strict(self):
        with FilterCtxStrict('age__in', (1, 2, 3)) as value:
            self.assertEqual(value, 'age IN (1,2,3)')

        with FilterCtxStrict('age__in', (1, '2', 3)) as value:
            self.assertEqual(value, 'age IN (1,2,3)')

        with FilterCtxStrict('age__in', []) as value:
            self.assertIsNone(value)

        with FilterCtxStrict('age__in', None) as value:
            self.assertIsNone(value)

        self.assertRaises(
            SphinxQLSyntaxException,
            lambda: FilterCtxStrict('age__in', [1, 2, 3, '42d']).__enter__()
        )

    def test_between_non_strict(self):
        with FilterCtxSoft('age__between', (1, 25)) as value:
            self.assertEqual(value, 'age BETWEEN 1 AND 25')

        with FilterCtxSoft('age__between', (1, 25, 35)) as value:
            self.assertIsNone(value)

        with FilterCtxSoft('age__between', (1, '25')) as value:
            self.assertEqual(value, 'age BETWEEN 1 AND 25')

        with FilterCtxSoft('age__between', (1, 'string')) as value:
            self.assertIsNone(value)

        with FilterCtxSoft('age__between', ()) as value:
            self.assertIsNone(value)

    def test_between_strict(self):
        self.assertRaises(
            SphinxQLSyntaxException,
            lambda: FilterCtxStrict('age__between', (1, 25, 35)).__enter__()
        )
        self.assertRaises(
            SphinxQLSyntaxException,
            lambda: FilterCtxStrict('age__between', (1, 'string')).__enter__()
        )
        with FilterCtxStrict('age__between', (1, 25)) as value:
            self.assertEqual(value, 'age BETWEEN 1 AND 25')

        with FilterCtxStrict('age__between', ()) as value:
            self.assertIsNone(value)

    def test_invalid_condition(self):
        with FilterCtxSoft('age__range', (1, 25)) as value:
            self.assertIsNone(value)

    def test_invalid_condition_strict(self):
        self.assertRaises(
            SphinxQLSyntaxException,
            lambda: FilterCtxStrict('age__range', (1, 25)).__enter__()
        )

    def test_datetime_condition(self):
        today_stamp = unix_timestamp(date.today())
        with FilterCtxStrict('age__eq', today_stamp) as value:
            self.assertEqual(value, 'age=%s' % today_stamp)

        now_stamp = unix_timestamp(datetime.now())
        with FilterCtxStrict('age__eq', now_stamp) as value:
            self.assertEqual(value, 'age=%s' % now_stamp)

    def test_float_condition(self):
        with FilterCtxStrict('age__eq', 5.5) as value:
            self.assertEqual(value, 'age=5.5')


MatchQueryCtxSoft = lambda *args, **kwargs: MatchQueryCtx(*args, **kwargs).with_config(ProductionConfig)
MatchQueryCtxStrict = lambda *args, **kwargs: MatchQueryCtx(*args, **kwargs).with_config(DebugConfig)

class TestMatchQueryCtx(unittest.TestCase):

    def test_single_slash(self):
        symbol = random.choice(ESCAPED_CHARS.single_escape)
        with MatchQueryCtxSoft(symbol) as value:
            self.assertEqual(value, r"\{0}".format(symbol))

    def test_double_slash(self):
        symbol = random.choice(ESCAPED_CHARS.double_escape)
        with MatchQueryCtxSoft(symbol) as value:
            self.assertEqual(value, r"\\{0}".format(symbol))

    def test_single_slash_raw(self):
        symbol = random.choice(ESCAPED_CHARS.single_escape)
        with MatchQueryCtxSoft(symbol, raw=True) as value:
            self.assertEqual(value, symbol)

    def test_raw_double_slash(self):
        symbol = random.choice(ESCAPED_CHARS.double_escape)
        with MatchQueryCtxSoft(symbol, raw=True) as value:
            self.assertEqual(value, symbol)

    def test_simple_valid_string(self):
        with MatchQueryCtxSoft('Find me some money') as value:
            self.assertEqual(value, "Find me some money")

    def test_symbol_escape(self):
        with MatchQueryCtxSoft('@full_name ^Roman') as value:
            self.assertEqual(value, r'\\@full_name \\^Roman')

    def test_quotes_escape(self):
        with MatchQueryCtxSoft('"name') as value:
            self.assertEqual(value, r'\\"name')

    def test_quote_escape(self):
        with MatchQueryCtxSoft("l'amour") as value:
            self.assertEqual(value, r"l\'amour")

    def test_raw_query(self):
        mail_query = '@email "semirook@gmail.com"'
        with MatchQueryCtxSoft(mail_query, raw=True) as value:
            self.assertEqual(value, mail_query)

    def test_invalid_query(self):
        with MatchQueryCtxSoft(42) as value:
            self.assertIsNone(value)

        with MatchQueryCtxSoft('') as value:
            self.assertIsNone(value)

        with MatchQueryCtxSoft('   ') as value:
            self.assertIsNone(value)

    def test_invalid_query_strict(self):
        self.assertRaises(
            SphinxQLSyntaxException,
            lambda: MatchQueryCtxStrict(42).__enter__()
        )
        with MatchQueryCtxStrict('') as value:
            self.assertIsNone(value)

        with MatchQueryCtxStrict('   ') as value:
            self.assertIsNone(value)


AliasFieldCtxSoft = lambda x, y: AliasFieldCtx(x, y).with_config(ProductionConfig)
AliasFieldCtxStrict = lambda x, y: AliasFieldCtx(x, y).with_config(DebugConfig)

class TestAliasFieldCtx(unittest.TestCase):

    def test_valid_attrs(self):
        with AliasFieldCtxSoft('name', 'another_name') as value:
            self.assertEqual(value, 'name AS another_name')

    def test_valid_attrs_strict(self):
        with AliasFieldCtxStrict('name', 'another_name') as value:
            self.assertEqual(value, 'name AS another_name')

    def test_invalid_attrs(self):
        with AliasFieldCtxSoft('name', 42) as value:
            self.assertIsNone(value)

        with AliasFieldCtxSoft('name', random.choice(RESERVED_KEYWORDS)) as value:
            self.assertIsNone(value)

        with AliasFieldCtxSoft(random.choice(RESERVED_KEYWORDS), 'alias') as value:
            self.assertIsNone(value)

        with AliasFieldCtxSoft('', '') as value:
            self.assertIsNone(value)

        with AliasFieldCtxSoft('   ', '   ') as value:
            self.assertIsNone(value)

    def test_invalid_attrs_strict(self):
        self.assertRaises(
            SphinxQLSyntaxException,
            lambda: AliasFieldCtxStrict('name', 42).__enter__()
        )
        self.assertRaises(
            SphinxQLSyntaxException,
            lambda: AliasFieldCtxStrict('name', random.choice(RESERVED_KEYWORDS)).__enter__()
        )
        self.assertRaises(
            SphinxQLSyntaxException,
            lambda: AliasFieldCtxStrict(random.choice(RESERVED_KEYWORDS), 'alias').__enter__()
        )
        self.assertRaises(
            SphinxQLSyntaxException,
            lambda: AliasFieldCtxStrict('', '').__enter__()
        )
        self.assertRaises(
            SphinxQLSyntaxException,
            lambda: AliasFieldCtxStrict('   ', '   ').__enter__()
        )


FieldCtxSoft = lambda field: FieldCtx(field).with_config(ProductionConfig)
FieldCtxStrict = lambda field: FieldCtx(field).with_config(DebugConfig)

class TestFieldCtx(unittest.TestCase):

    def test_valid_attrs(self):
        with FieldCtxSoft('name') as lex:
            self.assertEqual(lex, 'name')

    def test_valid_attrs_strict(self):
        with FieldCtxStrict('name') as lex:
            self.assertEqual(lex, 'name')

    def test_invalid_attrs(self):
        with FieldCtxSoft('') as lex:
            self.assertIsNone(lex)

        with FieldCtxSoft('   ') as lex:
            self.assertIsNone(lex)

        with FieldCtxSoft(random.choice(RESERVED_KEYWORDS)) as lex:
            self.assertIsNone(lex)

    def test_invalid_attrs_strict(self):
        self.assertRaises(
            SphinxQLSyntaxException,
            lambda: FieldCtxStrict('').__enter__()
        )
        self.assertRaises(
            SphinxQLSyntaxException,
            lambda: FieldCtxStrict('   ').__enter__()
        )
        self.assertRaises(
            SphinxQLSyntaxException,
            lambda: FieldCtxStrict(random.choice(RESERVED_KEYWORDS)).__enter__()
        )


LimitCtxSoft = lambda x, y: LimitCtx(x, y).with_config(ProductionConfig)
LimitCtxStrict = lambda x, y: LimitCtx(x, y).with_config(DebugConfig)

class TestLimitCtx(unittest.TestCase):

    def test_valid_attrs(self):
        with LimitCtxSoft(0, 100) as lex:
            self.assertEqual([0, 100], lex)

        with LimitCtxSoft(0, '100') as lex:
            self.assertEqual([0, 100], lex)

    def test_invalid_attrs(self):
        with LimitCtxSoft(0, 0) as lex:
            self.assertIsNone(lex)

        with LimitCtxSoft(-2, 100) as lex:
            self.assertIsNone(lex)

        with LimitCtxSoft(0, 'boom') as lex:
            self.assertIsNone(lex)

    def test_invalid_attrs_strict(self):
        self.assertRaises(
            SphinxQLSyntaxException,
            lambda: LimitCtxStrict(0, 0).__enter__()
        )
        self.assertRaises(
            SphinxQLSyntaxException,
            lambda: LimitCtxStrict(-2, 100).__enter__()
        )
        self.assertRaises(
            SphinxQLSyntaxException,
            lambda: LimitCtxStrict(0, 'boom').__enter__()
        )


OrderCtxSoft = lambda x, y: OrderCtx(x, y).with_config(ProductionConfig)
OrderCtxStrict = lambda x, y: OrderCtx(x, y).with_config(DebugConfig)

class TestOrderCtx(unittest.TestCase):

    def test_valid_attrs(self):
        with OrderCtxSoft('name', 'ASC') as lex:
            self.assertEqual(lex, 'name ASC')

        with OrderCtxSoft('name', 'DESC') as lex:
            self.assertEqual(lex, 'name DESC')

        with OrderCtxSoft('name', 'asc') as lex:
            self.assertEqual(lex, 'name ASC')

        with OrderCtxSoft('name', 'desc') as lex:
            self.assertEqual(lex, 'name DESC')

    def test_invalid_attrs(self):
        with OrderCtxSoft('', 'ASC') as lex:
            self.assertIsNone(lex)

        with OrderCtxSoft('   ', 'ASC') as lex:
            self.assertIsNone(lex)

        with OrderCtxSoft('name', 'DEC') as lex:
            self.assertIsNone(lex)

        with OrderCtxSoft(42, 'ASC') as lex:
            self.assertIsNone(lex)

        with OrderCtxSoft('asc', 'name') as lex:
            self.assertIsNone(lex)

    def test_invalid_attrs_strict(self):
        self.assertRaises(
            SphinxQLSyntaxException,
            lambda: OrderCtxStrict('', 'ASC').__enter__()
        )
        self.assertRaises(
            SphinxQLSyntaxException,
            lambda: OrderCtxStrict('   ', 'ASC').__enter__()
        )
        self.assertRaises(
            SphinxQLSyntaxException,
            lambda: OrderCtxStrict('name', 'DEC').__enter__()
        )
        self.assertRaises(
            SphinxQLSyntaxException,
            lambda: OrderCtxStrict(42, 'ASC').__enter__()
        )
        self.assertRaises(
            SphinxQLSyntaxException,
            lambda: OrderCtxStrict('asc', '').__enter__()
        )


OptionsCtxSoft = lambda x, y: OptionsCtx(x, y).with_config(ProductionConfig)
OptionsCtxStrict = lambda x, y: OptionsCtx(x, y).with_config(DebugConfig)

class TestOptionsCtx(unittest.TestCase):

    def test_get_ranker_valid_attrs(self):
        with OptionsCtxSoft('ranker', 'proximity_bm25') as lex:
            self.assertEqual(lex, 'ranker=proximity_bm25')

        with OptionsCtxSoft('ranker', 'bm25') as lex:
            self.assertEqual(lex, 'ranker=bm25')

        with OptionsCtxSoft('ranker', 'none') as lex:
            self.assertEqual(lex, 'ranker=none')

        with OptionsCtxSoft('ranker', 'wordcount') as lex:
            self.assertEqual(lex, 'ranker=wordcount')

        with OptionsCtxSoft('ranker', 'proximity') as lex:
            self.assertEqual(lex, 'ranker=proximity')

        with OptionsCtxSoft('ranker', 'matchany') as lex:
            self.assertEqual(lex, 'ranker=matchany')

        with OptionsCtxSoft('ranker', 'fieldmask') as lex:
            self.assertEqual(lex, 'ranker=fieldmask')

    def test_get_ranker_invalid_attrs(self):
        with OptionsCtxSoft('ranker', 'yandex') as lex:
            self.assertIsNone(lex)

    def test_get_ranker_invalid_attrs_strict(self):
        self.assertRaises(
            SphinxQLSyntaxException,
            lambda: OptionsCtxStrict('ranker', 'yandex').__enter__()
        )

    def test_get_max_matches_valid_attrs(self):
        with OptionsCtxSoft('max_matches', 3000) as lex:
            self.assertEqual(lex, 'max_matches=3000')

        with OptionsCtxSoft('max_matches', '3000') as lex:
            self.assertEqual(lex, 'max_matches=3000')

    def test_get_max_matches_invalid_attrs(self):
        with OptionsCtxSoft('max_matches', 0) as lex:
            self.assertIsNone(lex)

        with OptionsCtxSoft('max_matches', '') as lex:
            self.assertIsNone(lex)

        with OptionsCtxSoft('max_matches', 'many') as lex:
            self.assertIsNone(lex)

    def test_get_max_matches_invalid_attrs_strict(self):
        self.assertRaises(
            SphinxQLSyntaxException,
            lambda: OptionsCtxStrict('max_matches', 0).__enter__()
        )
        self.assertRaises(
            SphinxQLSyntaxException,
            lambda: OptionsCtxStrict('max_matches', '').__enter__()
        )
        self.assertRaises(
            SphinxQLSyntaxException,
            lambda: OptionsCtxStrict('max_matches', 'many').__enter__()
        )

    def test_get_cutoff_valid_attrs(self):
        with OptionsCtxSoft('cutoff', 7) as lex:
            self.assertEqual(lex, 'cutoff=7')

        with OptionsCtxSoft('cutoff', '7') as lex:
            self.assertEqual(lex, 'cutoff=7')

    def test_get_cutoff_invalid_attrs(self):
        with OptionsCtxSoft('cutoff', 0) as lex:
            self.assertIsNone(lex)

        with OptionsCtxSoft('cutoff', '') as lex:
            self.assertIsNone(lex)

        with OptionsCtxSoft('cutoff', 'many') as lex:
            self.assertIsNone(lex)

    def test_get_cutoff_invalid_attrs_strict(self):
        self.assertRaises(
            SphinxQLSyntaxException,
            lambda: OptionsCtxStrict('cutoff', 0).__enter__()
        )
        self.assertRaises(
            SphinxQLSyntaxException,
            lambda: OptionsCtxStrict('cutoff', '').__enter__()
        )
        self.assertRaises(
            SphinxQLSyntaxException,
            lambda: OptionsCtxStrict('cutoff', 'many').__enter__()
        )

    def test_max_query_time_valid_attrs(self):
        with OptionsCtxSoft('max_query_time', 100) as lex:
            self.assertEqual(lex, 'max_query_time=100')

        with OptionsCtxSoft('max_query_time', '100') as lex:
            self.assertEqual(lex, 'max_query_time=100')

    def test_max_query_time_invalid_attrs(self):
        with OptionsCtxSoft('max_query_time', '') as lex:
            self.assertIsNone(lex)

        with OptionsCtxSoft('max_query_time', 0) as lex:
            self.assertIsNone(lex)

    def test_max_query_time_invalid_attrs_strict(self):
        self.assertRaises(
            SphinxQLSyntaxException,
            lambda: OptionsCtxStrict('max_query_time', '').__enter__()
        )
        self.assertRaises(
            SphinxQLSyntaxException,
            lambda: OptionsCtxStrict('max_query_time', 0).__enter__()
        )

    def test_retry_count_valid_attrs(self):
        with OptionsCtxSoft('retry_count', 3) as lex:
            self.assertEqual(lex, 'retry_count=3')

        with OptionsCtxSoft('retry_count', '3') as lex:
            self.assertEqual(lex, 'retry_count=3')

    def test_retry_count_invalid_attrs(self):
        with OptionsCtxSoft('retry_count', '') as lex:
            self.assertIsNone(lex)

        with OptionsCtxSoft('retry_count', 0) as lex:
            self.assertIsNone(lex)

    def test_retry_count_invalid_attrs_strict(self):
        self.assertRaises(
            SphinxQLSyntaxException,
            lambda: OptionsCtxStrict('retry_count', '').__enter__()
        )
        self.assertRaises(
            SphinxQLSyntaxException,
            lambda: OptionsCtxStrict('retry_count', 0).__enter__()
        )

    def test_retry_delay_valid_attrs(self):
        with OptionsCtxSoft('retry_delay', 3) as lex:
            self.assertEqual(lex, 'retry_delay=3')

        with OptionsCtxSoft('retry_delay', '3') as lex:
            self.assertEqual(lex, 'retry_delay=3')

    def test_retry_delay_invalid_attrs(self):
        with OptionsCtxSoft('retry_delay', '') as lex:
            self.assertIsNone(lex)

        with OptionsCtxSoft('retry_delay', 0) as lex:
            self.assertIsNone(lex)

    def test_retry_delay_invalid_attrs_strict(self):
        self.assertRaises(
            SphinxQLSyntaxException,
            lambda: OptionsCtxStrict('retry_delay', '').__enter__()
        )
        self.assertRaises(
            SphinxQLSyntaxException,
            lambda: OptionsCtxStrict('retry_delay', 0).__enter__()
        )

    def test_field_weights_valid_attrs(self):
        with OptionsCtxSoft('field_weights', {'title': 100, 'body': 50}) as lex:
            self.assertTrue(lex in [
                'field_weights=(body=50, title=100)',
                'field_weights=(title=100, body=50)'
            ])
        with OptionsCtxSoft('field_weights', {'title': '100', 'body': '50'}) as lex:
            self.assertTrue(lex in [
                'field_weights=(body=50, title=100)',
                'field_weights=(title=100, body=50)'
            ])

    def test_field_weights_invalid_attrs(self):
        with OptionsCtxSoft('field_weights', ('title', 0)) as lex:
            self.assertIsNone(lex)

        with OptionsCtxSoft('field_weights', 100) as lex:
            self.assertIsNone(lex)

        with OptionsCtxSoft('field_weights', {'title': 10, 'body': 0}) as lex:
            self.assertEqual(lex, 'field_weights=(title=10)')

    def test_field_weights_invalid_attrs_strict(self):
        self.assertRaises(
            SphinxQLSyntaxException,
            lambda: OptionsCtxStrict('field_weights', ('title', 0)).__enter__()
        )
        self.assertRaises(
            SphinxQLSyntaxException,
            lambda: OptionsCtxStrict('field_weights', 100).__enter__()
        )
        self.assertRaises(
            SphinxQLSyntaxException,
            lambda: OptionsCtxStrict('field_weights', {'title': 10, 'body': 0}).__enter__()
        )

    def test_index_weights_valid_attrs(self):
        with OptionsCtxSoft('index_weights', {'company': 1, 'company_delta': 2}) as lex:
            self.assertTrue(lex in [
                'index_weights=(company=1, company_delta=2)',
                'index_weights=(company_delta=2, company=1)'
            ])
        with OptionsCtxSoft('index_weights', {'company': '1', 'company_delta': '2'}) as lex:
            self.assertTrue(lex in [
                'index_weights=(company=1, company_delta=2)',
                'index_weights=(company_delta=2, company=1)'
            ])

    def test_index_weights_invalid_attrs(self):
        with OptionsCtxSoft('index_weights', ('company', 0)) as lex:
            self.assertIsNone(lex)

        with OptionsCtxSoft('index_weights', 100) as lex:
            self.assertIsNone(lex)

        with OptionsCtxSoft('index_weights', {'company': 1, 'company_delta': 0}) as lex:
            self.assertEqual(lex, 'index_weights=(company=1)')

    def test_index_weights_invalid_attrs_strict(self):
        self.assertRaises(
            SphinxQLSyntaxException,
            lambda: OptionsCtxStrict('index_weights', ('company', 0)).__enter__()
        )
        self.assertRaises(
            SphinxQLSyntaxException,
            lambda: OptionsCtxStrict('index_weights', 100).__enter__()
        )
        self.assertRaises(
            SphinxQLSyntaxException,
            lambda: OptionsCtxStrict('index_weights', {'company': 1, 'company_delta': 0}).__enter__()
        )

    def test_reverse_scan_valid_attrs(self):
        with OptionsCtxSoft('reverse_scan', True) as lex:
            self.assertEqual(lex, 'reverse_scan=1')

        with OptionsCtxSoft('reverse_scan', False) as lex:
            self.assertEqual(lex, 'reverse_scan=0')

    def test_reverse_scan_invalid_attrs(self):
        with OptionsCtxSoft('reverse_scan', 1) as lex:
            self.assertIsNone(lex)

        with OptionsCtxSoft('reverse_scan', 0) as lex:
            self.assertIsNone(lex)

        with OptionsCtxSoft('reverse_scan', '1') as lex:
            self.assertIsNone(lex)

    def test_reverse_scan_invalid_attrs_strict(self):
        self.assertRaises(
            SphinxQLSyntaxException,
            lambda: OptionsCtxStrict('reverse_scan', 1).__enter__()
        )
        self.assertRaises(
            SphinxQLSyntaxException,
            lambda: OptionsCtxStrict('reverse_scan', 0).__enter__()
        )
        self.assertRaises(
            SphinxQLSyntaxException,
            lambda: OptionsCtxStrict('reverse_scan', '0').__enter__()
        )

    def test_comment_valid_attrs(self):
        with OptionsCtxSoft('comment', 'Comment text') as lex:
            self.assertEqual(lex, 'comment=Comment text')

    def test_comment_invalid_attrs(self):
        with OptionsCtxSoft('comment', 42) as lex:
            self.assertIsNone(lex)

    def test_comment_invalid_attrs_strict(self):
        self.assertRaises(
            SphinxQLSyntaxException,
            lambda: OptionsCtxStrict('comment', 42).__enter__()
        )


UpdateSetCtxSoft = lambda x, y: UpdateSetCtx(x, y).with_config(ProductionConfig)
UpdateSetCtxStrict = lambda x, y: UpdateSetCtx(x, y).with_config(DebugConfig)

class TestUpdateSetCtx(unittest.TestCase):

    def test_valid_attrs(self):
        with UpdateSetCtxSoft('intattr', 1000) as lex:
            self.assertEqual(lex, 'intattr=1000')

        with UpdateSetCtxSoft('intattr', '1000') as lex:
            self.assertEqual(lex, 'intattr=1000')

        with UpdateSetCtxSoft('fattr', 100.002) as lex:
            self.assertEqual(lex, 'fattr=100.002')

        with UpdateSetCtxSoft('mvattr', [1, 2, 3]) as lex:
            self.assertEqual(lex, 'mvattr=(1,2,3)')

        with UpdateSetCtxSoft('mvattr', [1, '2', 3]) as lex:
            self.assertEqual(lex, 'mvattr=(1,2,3)')

        with UpdateSetCtxSoft('mvattr', (1, '2', 3)) as lex:
            self.assertEqual(lex, 'mvattr=(1,2,3)')

        with UpdateSetCtxSoft('attr', '') as lex:
            self.assertEqual(lex, 'attr=()')

        with UpdateSetCtxSoft('attr', []) as lex:
            self.assertEqual(lex, 'attr=()')

        with UpdateSetCtxSoft('attr', ()) as lex:
            self.assertEqual(lex, 'attr=()')

        with UpdateSetCtxSoft('attr', None) as lex:
            self.assertEqual(lex, 'attr=()')

    def test_invalid_attrs(self):
        with UpdateSetCtxSoft('', 1000) as lex:
            self.assertIsNone(lex)

        with UpdateSetCtxSoft('attr', 'string_is_not_supported') as lex:
            self.assertIsNone(lex)

        with UpdateSetCtxSoft('attr', ['1', 2, None, 'string']) as lex:
            self.assertEqual(lex, 'attr=(1,2)')



"""
    sphinxit.core.exceptions
    ~~~~~~~~~~~~~~~~~~~~~~~~

    Implements Sphinxit typical exceptions.

    :copyright: (c) 2013 by Roman Semirook.
    :license: BSD, see LICENSE for more details.
"""


class SphinxQLDriverException(Exception):
    pass


class SphinxQLSyntaxException(Exception):
    pass


class SphinxQLChainException(Exception):
    pass


class ImproperlyConfigured(Exception):
    pass



"""
    sphinxit.core.mixins
    ~~~~~~~~~~~~~~~~~~~~

    Implements mixins for nodes and processor.

    :copyright: (c) 2013 by Roman Semirook.
    :license: BSD, see LICENSE for more details.
"""

from __future__ import unicode_literals

from .exceptions import ImproperlyConfigured


class MagicMixin(object):

    def __bool__(self):
        return True

    def __nonzero__(self):
        return self.__bool__()

    def __str__(self):
        return self.lex()

    def lex(self):
        raise NotImplementedError("It's just mixin")


class ConfigMixin(MagicMixin):

    def __init__(self):
        self._config = None

    @property
    def config(self):
        if self._config is None:
            raise ImproperlyConfigured(
                'No config provided for instance processing'
            )
        return self._config

    @config.setter
    def config(self, value):
        self._config = value

    def with_config(self, config):
        self._config = config
        return self

    def has_config(self):
        return self._config is not None

    @property
    def is_strict(self):
        return getattr(self.config, 'DEBUG', False)


class CtxMixin(ConfigMixin):

    def __exit__(self, exc_type=None, exc_val=None, exc_tb=None):
        if exc_val is not None and self.is_strict:
            raise exc_val
        return None



from setuptools import setup, find_packages


setup(
    name='sphinxit',
    version='0.3.2',
    author='Roman Semirook',
    author_email='semirook@gmail.com',
    packages=find_packages(),
    license='BSD',
    url='https://github.com/semirook/sphinxit',
    description='Lite and powerful SphinxQL query constructor',
    long_description='Lite and powerful SphinxQL query constructor',
    install_requires=[
        "six >= 1.1.0",
        "oursql >= 0.9.3",
        "ordereddict >= 1.1",
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 2 :: Only',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)



# coding=utf-8
from __future__ import unicode_literals
import datetime

try:
    import unittest2 as unittest
except ImportError:
    import unittest

from sphinxit.core.nodes import Count, OR, RawAttr
from sphinxit.core.processor import Search, Snippet
from sphinxit.core.helpers import unix_timestamp, BaseSearchConfig


class SearchConfig(BaseSearchConfig):
    WITH_STATUS = False


class TestSearch(unittest.TestCase):

    def test_simple(self):
        search = Search(indexes=['company'], config=SearchConfig)
        search = search.match('Yandex')
        self.assertEqual(
            search.lex(),
            "SELECT * FROM company WHERE MATCH('Yandex')"
        )

    def test_mutability(self):
        search = Search(indexes=['company'], config=SearchConfig)
        search.match('Yandex')
        self.assertEqual(
            search.lex(),
            "SELECT * FROM company"
        )
        search = search.match('Yandex')
        self.assertEqual(
            search.lex(),
            "SELECT * FROM company WHERE MATCH('Yandex')"
        )

    def test_with_select(self):
        search = Search(indexes=['company'], config=SearchConfig)
        search = search.select('id', 'date_created')
        search = search.match('Yandex')
        self.assertEqual(
            search.lex(),
            "SELECT id, date_created FROM company WHERE MATCH('Yandex')"
        )

    def test_with_or_filters(self):
        correct_qls = [
            "SELECT *, (id>=100 OR id=1) AS cnd FROM company WHERE MATCH('Yandex') AND cnd>0",
            "SELECT *, (id=1 OR id>=100) AS cnd FROM company WHERE MATCH('Yandex') AND cnd>0",
        ]
        search = Search(['company'], config=SearchConfig)
        search = search.match('Yandex').filter(OR(id__gte=100, id__eq=1))
        self.assertIn(search.lex(), correct_qls)

    def test_with_or_filters_and_fields(self):
        correct_qls = [
            "SELECT id, (id>=100 OR id=1) AS cnd FROM company WHERE MATCH('Yandex') AND cnd>0",
            "SELECT id, (id=1 OR id>=0) AS cnd FROM company WHERE MATCH('Yandex') AND cnd>0",
        ]
        search = Search(['company'], config=SearchConfig).select('id')
        search = search.match('Yandex').filter(OR(id__gte=100, id__eq=1))
        self.assertIn(search.lex(), correct_qls)

    def test_with_params(self):
        search = Search(['company'], config=SearchConfig)
        search = search.match('Yandex').limit(0, 100).order_by('name', 'desc')
        self.assertEqual(
            search.lex(),
            "SELECT * FROM company WHERE MATCH('Yandex') ORDER BY name DESC LIMIT 0,100"
        )

    def test_with_options(self):
        search = Search(['company'], config=SearchConfig)
        search = (
            search
            .match('Yandex')
            .select('id', 'name')
            .options(
                ranker='proximity',
                max_matches=100,
                field_weights={'name': 100},
            )
            .order_by('name', 'desc')
        )
        correct_options_qls = [
            "max_matches=100, ranker=proximity, field_weights=(name=100)",
            "max_matches=100, field_weights=(name=100), ranker=proximity",
            "ranker=proximity, field_weights=(name=100), max_matches=100",
            "ranker=proximity, max_matches=100, field_weights=(name=100)",
            "field_weights=(name=100), max_matches=100, ranker=proximity",
            "field_weights=(name=100), ranker=proximity, max_matches=100",
        ]
        correct_qls = [
            " ".join((
                "SELECT id, name FROM company WHERE MATCH('Yandex') ORDER BY name DESC OPTION",
                opt
            ))
            for opt in correct_options_qls
        ]
        self.assertIn(search.lex(), correct_qls)

    def test_with_double_match(self):
        search = Search(['company'], config=SearchConfig)
        search = search.match('ОАО').match('ТНК')
        self.assertEqual(
            search.lex(),
            "SELECT * FROM company WHERE MATCH('ОАО ТНК')"
        )

    def test_with_time_filter(self):
        search = Search(['company'], config=SearchConfig)
        search = search.match('Yandex').filter(date_created__lte=datetime.date.today())
        today = datetime.date.today()
        sxql = (
            "SELECT * FROM company WHERE MATCH('Yandex') "
            "AND date_created<=%s" % unix_timestamp(today)
        )
        self.assertEqual(search.lex(), sxql)

    def test_with_raw_attr(self):
        search = Search(['company'], config=SearchConfig)
        search = search.match('Yandex').select(RawAttr('@weight*10', 'skey'))
        self.assertEqual(
            search.lex(),
            "SELECT @weight*10 AS skey FROM company WHERE MATCH('Yandex')"
        )

    def test_update_syntax(self):
        search = Search(['company'], config=SearchConfig)
        search = search.match('Yandex').update(products=(5,2)).filter(id__gt=1)
        self.assertEqual(
            search.lex(),
            "UPDATE company SET products=(5,2) WHERE MATCH('Yandex') AND id>1"
        )

    def test_with_grouping(self):
        search = Search(['company'], config=SearchConfig)
        search = search.match('Yandex').select(Count()).group_by('date_created')
        self.assertEqual(
            search.lex(),
            "SELECT COUNT(*) AS num FROM company WHERE MATCH('Yandex') GROUP BY date_created"
        )

    def test_with_multiindex(self):
        search = Search(
            indexes=['company', 'company_delta'],
            config=SearchConfig
        )
        search = search.filter(id__gte=100)
        self.assertEqual(
            search.lex(),
            "SELECT * FROM company, company_delta WHERE id>=100"
        )


class TestSnippets(unittest.TestCase):

    def test_simple(self):
        snippets = (
            Snippet(index='company', config=SearchConfig)
            .for_query("Me amore")
            .from_data("amore")
        )
        self.assertEqual(
            snippets.lex(),
            "CALL SNIPPETS ('amore', 'company', 'Me amore')"
        )

    def test_extended_1(self):
        snippets = (
            Snippet(index='company', config=SearchConfig)
            .for_query("Me amore")
            .from_data("amore", "amore mia")
        )
        self.assertEqual(
            snippets.lex(),
            "CALL SNIPPETS (('amore', 'amore mia'), 'company', 'Me amore')"
        )

    def test_extended_2(self):
        snippets = (
            Snippet(index='company', config=SearchConfig)
            .for_query("Me amore")
            .from_data("amore")
            .from_data("me amore")
        )
        self.assertEqual(
            snippets.lex(),
            "CALL SNIPPETS (('amore', 'me amore'), 'company', 'Me amore')"
        )

    def test_with_options(self):
        snippets = (
            Snippet(index='company', config=SearchConfig)
            .for_query("Me amore")
            .from_data("amore mia")
            .options(before_match='<strong>', after_match='</strong>')
        )
        self.assertEqual(
            snippets.lex(), (
                "CALL SNIPPETS ('amore mia', 'company', 'Me amore', "
                "'<strong>' AS before_match, '</strong>' AS after_match)"
            )
        )



"""
    sphinxit.core.convertors
    ~~~~~~~~~~~~~~~~~~~~~~~~

    Implements convertors and cleaners processors.

    :copyright: (c) 2013 by Roman Semirook.
    :license: BSD, see LICENSE for more details.
"""

from __future__ import unicode_literals

import re
import six
from datetime import datetime, date

from sphinxit.core.constants import ESCAPED_CHARS, RESERVED_KEYWORDS
from sphinxit.core.exceptions import SphinxQLSyntaxException
from sphinxit.core.helpers import (
    list_of_integers_only,
    int_from_digit,
    unix_timestamp
)
from sphinxit.core.mixins import CtxMixin


class FilterCtx(CtxMixin):
    _allowed_conditions_map = {
        '__eq': '{a}={v}',
        '__neq': '{a}!={v}',
        '__gt': '{a}>{v}',
        '__gte': '{a}>={v}',
        '__lt': '{a}<{v}',
        '__lte': '{a}<={v}',
        '__in': '{a} IN ({v})',
        '__between': '{a} BETWEEN {f_v} AND {s_v}',
    }

    def __init__(self, k_attr, v_attr):
        super(FilterCtx, self).__init__()
        self.k_attr = k_attr
        self.v_attr = v_attr

    def __enter__(self):
        v_attr = self.v_attr
        if isinstance(v_attr, six.string_types):
            v_attr = int_from_digit(
                v_attr,
                is_strict=self.is_strict
            )
        if isinstance(self.v_attr, (tuple, list)):
            v_attr = list_of_integers_only(
                v_attr,
                is_strict=self.is_strict
            )
        if isinstance(self.v_attr, (datetime, date)):
            v_attr = unix_timestamp(self.v_attr)

        if not v_attr and v_attr != 0:
            return None

        for ending in self._allowed_conditions_map.keys():
            if not self.k_attr.endswith(ending):
                continue

            if (
                ending not in ('__between', '__in')
                and isinstance(v_attr, (tuple, list))
                and not self.__exit__(
                    exc_val=SphinxQLSyntaxException(
                        '%s found but not allowed for %s condition' %
                        (self.v_attr, self.k_attr)
                    )
                )
            ):
                continue

            if (
                ending in ('__between', '__in')
                and not isinstance(v_attr, (tuple, list))
                and not self.__exit__(
                    exc_val=SphinxQLSyntaxException(
                        '%s condition found but the type of %s is not list or tuple' %
                        (self.k_attr, self.v_attr)
                    )
                )
            ):
                continue

            a = self.k_attr[:self.k_attr.rindex(ending)]
            v = v_attr

            if ending == '__between':
                if (
                    (len(v_attr) != 2 or len(v_attr) != len(self.v_attr))
                    and not self.__exit__(
                        exc_val=SphinxQLSyntaxException(
                            '%s condition wants a pair value and %s is not' %
                            (self.k_attr, self.v_attr)
                        )
                    )
                ):
                    continue

                f_v, s_v = v_attr
                return self._allowed_conditions_map[ending].format(
                    a=a,
                    f_v=f_v,
                    s_v=s_v
                )

            if ending == '__in':
                v = ','.join([str(v) for v in v_attr])

            return self._allowed_conditions_map[ending].format(a=a, v=v)

        return self.__exit__(
            exc_val=SphinxQLSyntaxException(
                '%s is invalid condition' % self.k_attr
            )
        )


class ORFilterCtx(FilterCtx):
    _allowed_conditions_map = {
        '__eq': '{a}={v}',
        '__gt': '{a}>{v}',
        '__gte': '{a}>={v}',
        '__lt': '{a}<{v}',
        '__lte': '{a}<={v}',
    }


class MatchQueryCtx(CtxMixin):

    def __init__(self, query, raw=False):
        super(MatchQueryCtx, self).__init__()
        self.query = query
        self.is_raw = raw

    def __enter__(self):
        if not isinstance(self.query, six.string_types):
            return self.__exit__(
                exc_val=SphinxQLSyntaxException(
                    '"%s" query is not a string' % self.query
                )
            )
        if not bool(self.query.strip()):
            return None

        if not self.is_raw:
            single_escape_chars_re = '|\\'.join(ESCAPED_CHARS.single_escape)
            self.query = re.sub(
                single_escape_chars_re,
                lambda m: r'\%s' % m.group(),
                self.query
            )
            double_escape_chars_re = '|\\'.join(ESCAPED_CHARS.double_escape)
            self.query = re.sub(
                double_escape_chars_re,
                lambda m: r'\\%s' % m.group(),
                self.query
            )

        return self.query


class FieldCtx(CtxMixin):

    def __init__(self, field):
        super(FieldCtx, self).__init__()
        self.field = field

    def __enter__(self):
        if not isinstance(self.field, six.string_types):
            return self.__exit__(
                exc_val=SphinxQLSyntaxException(
                    '"%s" is not a string' % self.field
                )
            )
        if self.field.upper() in RESERVED_KEYWORDS:
            return self.__exit__(
                exc_val=SphinxQLSyntaxException(
                    '"%s" is reserved keyword for Sphinx' % self.field
                )
            )
        if not bool(self.field.strip()):
            return self.__exit__(
                exc_val=SphinxQLSyntaxException('The field is empty')
            )

        return self.field


class AliasFieldCtx(CtxMixin):

    def __init__(self, field, alias):
        super(AliasFieldCtx, self).__init__()
        self.field = field
        self.alias = alias
        self.called_cls = None

    def called_by(self, cls):
        self.called_cls = cls
        return self

    def __enter__(self):
        error_prefix = (
            'Trouble with %s. ' % self.called_cls.__name__
            if self.called_cls else ''
        )
        if not isinstance(self.field, six.string_types):
            return self.__exit__(
                exc_val=SphinxQLSyntaxException(
                    error_prefix + '"%s" field is not a string' % self.field
                )
            )
        if not isinstance(self.alias, six.string_types):
            return self.__exit__(
                exc_val=SphinxQLSyntaxException(
                    '"%s" alias is not a string' % self.field
                )
            )
        if self.alias.upper() in RESERVED_KEYWORDS:
            return self.__exit__(
                exc_val=SphinxQLSyntaxException(
                    '"%s" is reserved keyword for Sphinx' % self.alias
                )
            )
        if self.field.upper() in RESERVED_KEYWORDS:
            return self.__exit__(
                exc_val=SphinxQLSyntaxException(
                    '"%s" is reserved keyword for Sphinx' % self.field
                )
            )
        if not bool(self.field.strip()):
            return self.__exit__(
                exc_val=SphinxQLSyntaxException('The field is empty')
            )
        if not bool(self.alias.strip()):
            return self.__exit__(
                exc_val=SphinxQLSyntaxException('The alias is empty')
            )

        return '%s AS %s' % (self.field, self.alias)


class OrderCtx(CtxMixin):

    def __init__(self, field, direction):
        super(OrderCtx, self).__init__()
        self.field = field
        self.direction = direction

    def __enter__(self):
        if not isinstance(self.field, six.string_types):
            return self.__exit__(
                exc_val=SphinxQLSyntaxException(
                    '"%s" field is not a string' % self.field
                )
            )
        if self.field.upper() in RESERVED_KEYWORDS:
            return self.__exit__(
                exc_val=SphinxQLSyntaxException(
                    '"%s" field is reserved keyword in Sphinx '
                    'and invalid for field name' % self.field
                )
            )
        if not bool(self.field.strip()):
            return self.__exit__(
                exc_val=SphinxQLSyntaxException('The field is empty')
            )
        if (
            not isinstance(self.direction, six.string_types)
            or self.direction.upper() not in ('ASC', 'DESC')
        ):
            return self.__exit__(
                exc_val=SphinxQLSyntaxException(
                    'Order direction can be ASC or DESC, "%s" is not' % self.direction
                )
            )

        return '%s %s' % (self.field, self.direction.upper())


class LimitCtx(CtxMixin):

    def __init__(self, offset, limit):
        super(LimitCtx, self).__init__()
        self.offset = offset
        self.limit = limit

    def __enter__(self):
        pair = list_of_integers_only(
            [self.offset, self.limit],
            is_strict=self.is_strict,
        )
        if len(pair) != 2:
            return self.__exit__(
                exc_val=SphinxQLSyntaxException(
                    'LIMIT clause wants a pair value and %s is not' % pair
                )
            )

        offset, limit = pair
        if offset < 0:
            return self.__exit__(
                exc_val=SphinxQLSyntaxException(
                    '%s offset is less then 0' % offset
                )
            )
        if limit <= 0:
            return self.__exit__(
                exc_val=SphinxQLSyntaxException(
                    'The limit value has to be greater then 0, %s is not' % limit
                )
            )

        return pair


class OptionsCtx(CtxMixin):

    def __init__(self, option, params):
        super(OptionsCtx, self).__init__()
        self.option = option
        self.params = params

    def __enter__(self):
        options_list = (
            'ranker',
            'max_matches',
            'cutoff',
            'max_query_time',
            'retry_count',
            'retry_delay',
            'field_weights',
            'index_weights',
            'reverse_scan',
            'comment',
        )
        if self.option in options_list:
            return getattr(self, 'get_%s' % self.option)()
        else:
            return self.__exit__(
                exc_val=SphinxQLSyntaxException(
                    '%s is unknown option' % self.option
                )
            )

    def get_ranker(self):
        ranker = self.params
        valid_rankers = (
            'proximity_bm25',
            'bm25',
            'none',
            'wordcount',
            'proximity',
            'matchany',
            'fieldmask',
            'sph04',
            'expr',
        )
        if not ranker in valid_rankers:
            return self.__exit__(
                exc_val=SphinxQLSyntaxException(
                    '%s is unknown ranker. '
                    'Valid values are %s' % (
                        ranker,
                        ', '.join(['"%s"' % r for r in valid_rankers])
                    )
                )
            )

        return 'ranker=%s' % ranker

    def get_max_matches(self):
        value = int_from_digit(self.params, is_strict=self.is_strict)
        if not value:
            return self.__exit__(
                exc_val=SphinxQLSyntaxException('"max_matches" option is empty')
            )
        return 'max_matches=%s' % value

    def get_cutoff(self):
        value = int_from_digit(self.params, is_strict=self.is_strict)
        if not value:
            return self.__exit__(
                exc_val=SphinxQLSyntaxException('"cutoff" option is empty')
            )
        return 'cutoff=%s' % value

    def get_max_query_time(self):
        value = int_from_digit(self.params, is_strict=self.is_strict)
        if not value:
            return self.__exit__(
                exc_val=SphinxQLSyntaxException('"max_query_time" option is empty')
            )
        return 'max_query_time=%s' % value

    def get_retry_count(self):
        value = int_from_digit(self.params, is_strict=self.is_strict)
        if not value:
            return self.__exit__(
                exc_val=SphinxQLSyntaxException('"retry_count" option is empty')
            )
        return 'retry_count=%s' % value

    def get_retry_delay(self):
        value = int_from_digit(self.params, is_strict=self.is_strict)
        if not value:
            return self.__exit__(
                exc_val=SphinxQLSyntaxException('"retry_delay" option is empty')
            )
        return 'retry_delay=%s' % value

    def get_field_weights(self):
        if not isinstance(self.params, dict):
            return self.__exit__(
                exc_val=SphinxQLSyntaxException(
                    '%s is not a dict as expected' % str(self.params)
                )
            )

        clean_pairs = []
        for field, value in self.params.items():
            if not isinstance(field, six.string_types):
                return self.__exit__(
                    exc_val=SphinxQLSyntaxException(
                        '%s is not a string as expected' % field
                    )
                )
            value = int_from_digit(value, is_strict=self.is_strict)
            if not value:
                if not self.__exit__(
                    exc_val=SphinxQLSyntaxException(
                        'One of the "field_weights" is not valid integer value'
                    )
                ):
                    continue
            else:
                clean_pairs.append('='.join([field, str(value)]))

        return 'field_weights=(%s)' % ', '.join(clean_pairs)

    def get_index_weights(self):
        if not isinstance(self.params, dict):
            return self.__exit__(
                exc_val=SphinxQLSyntaxException(
                    '%s is not a dict as expected' % str(self.params)
                )
            )

        clean_pairs = []
        for field, value in self.params.items():
            if not isinstance(field, six.string_types):
                return self.__exit__(
                    exc_val=SphinxQLSyntaxException(
                        '%s is not a string as expected' % field
                    )
                )
            value = int_from_digit(value, self.is_strict)
            if not value:
                if not self.__exit__(
                    exc_val=SphinxQLSyntaxException(
                        'One of the "index_weights" is not valid integer value'
                    )
                ):
                    continue
            else:
                clean_pairs.append('='.join([field, str(value)]))

        return 'index_weights=(%s)' % ', '.join(clean_pairs)

    def get_reverse_scan(self):
        if not isinstance(self.params, bool):
            return self.__exit__(
                exc_val=SphinxQLSyntaxException(
                    '%s has to be True or False' % self.params
                )
            )

        to_reverse = 1 if self.params else 0

        return 'reverse_scan=%s' % to_reverse

    def get_comment(self):
        if not isinstance(self.params, six.string_types):
            return self.__exit__(
                exc_val=SphinxQLSyntaxException(
                    '%s is not a string as expected' % self.params
                )
            )

        return 'comment=%s' % self.params


class UpdateSetCtx(CtxMixin):

    def __init__(self, k_attr, v_attr):
        super(UpdateSetCtx, self).__init__()
        self.k_attr = k_attr
        self.v_attr = v_attr

    def __enter__(self):
        if not isinstance(self.k_attr, six.string_types):
            return self.__exit__(
                exc_val=SphinxQLSyntaxException(
                    '%s is not a string as expected' % self.k_attr
                )
            )
        if not bool(self.k_attr.strip()):
            return self.__exit__(
                exc_val=SphinxQLSyntaxException(
                    'Field is empty'
                )
            )

        if isinstance(self.v_attr, (list, tuple)):
            v_attr = '(%s)' % ','.join([
                str(v) for v in list_of_integers_only(
                    self.v_attr,
                    self.is_strict,
                )
            ])
        elif (
            isinstance(self.v_attr, six.string_types)
            and self.v_attr.isdigit()
        ):
            v_attr = self.v_attr
        elif isinstance(self.v_attr, (six.integer_types, float)):
            v_attr = str(self.v_attr)
        elif (
            isinstance(self.v_attr, six.string_types)
            and not bool(self.v_attr.strip())
            or self.v_attr is None
        ):
            v_attr = '()'
        else:
            return self.__exit__(
                exc_val=SphinxQLSyntaxException(
                    '%s is improper value for UPDATE clause' % self.v_attr
                )
            )

        return '%s=%s' % (self.k_attr, v_attr)


class SnippetsOptionsCtx(CtxMixin):

    def __init__(self, option, params):
        super(SnippetsOptionsCtx, self).__init__()
        self.option = option
        self.params = params

    def __enter__(self):
        options_list = (
            'before_match',
            'after_match',
            'chunk_separator',
            'limit',
            'around',
            'exact_phrase',
            'use_boundaries',
            'weight_order',
            'query_mode',
            'force_all_words',
            'limit_passages',
            'limit_words',
            'start_passage_id',
            'load_files',
            'load_files_scattered',
            'html_strip_mode',
            'allow_empty',
            'passage_boundary',
            'emit_zones',
        )

        if self.option in options_list:
            return getattr(self, self.option)()
        else:
            return self.__exit__(
                exc_val=SphinxQLSyntaxException(
                    '"%s" is unknown option for the SNIPPET clause' % self.option
                )
            )

    def before_match(self):
        return "'%s' AS before_match" % self.params

    def after_match(self):
        return "'%s' AS after_match" % self.params

    def chunk_separator(self):
        return "'%s' AS chunk_separator" % self.params

    def limit(self):
        return "%s AS limit" % self.params

    def around(self):
        return "%s AS around" % self.params

    def exact_phrase(self):
        return "%s AS around" % self.params

    def use_boundaries(self):
        return "%s AS exact_phrase" % self.params

    def weight_order(self):
        return "%s AS weight_order" % self.params

    def query_mode(self):
        return "%s AS query_mode" % self.params

    def force_all_words(self):
        return "%s AS force_all_words" % self.params

    def limit_passages(self):
        return "%s AS limit_passages" % self.params

    def limit_words(self):
        return "%s AS limit_words" % self.params

    def start_passage_id(self):
        return "%s AS start_passage_id" % self.params

    def load_files(self):
        return "%s AS load_files" % self.params

    def load_files_scattered(self):
        return "%s AS load_files_scattered" % self.params

    def html_strip_mode(self):
        allowed_modes = ("none", "strip", "index", "retain")
        if self.params is None:
            self.params = "none"
        if self.params not in allowed_modes:
            return self.__exit__(
                exc_val=SphinxQLSyntaxException(
                    '"%s" is unknown strip mode' % self.params
                )
            )

        return "'%s' AS html_strip_mode" % self.params

    def allow_empty(self):
        return "%s AS allow_empty" % self.params

    def passage_boundary(self):
        allowed_modes = ("sentence", "paragraph", "zone")
        if self.params not in allowed_modes:
            return self.__exit__(
                exc_val=SphinxQLSyntaxException(
                    '"%s" is unknown boundary mode' % self.params
                )
            )

        return "'%s' AS passage_boundary" % self.params

    def emit_zones(self):
        return "%s AS emit_zones" % self.params



"""
    sphinxit.core.helpers
    ~~~~~~~~~~~~~~~~~~~~~

    Implements useful set of helpers.

    :copyright: (c) 2013 by Roman Semirook.
    :license: BSD, see LICENSE for more details.
"""

from __future__ import unicode_literals

import time
import six

from sphinxit.core.exceptions import SphinxQLSyntaxException


def int_from_digit(value, is_strict=False):
    try:
        return int(value)
    except (ValueError, TypeError):
        if is_strict:
            raise SphinxQLSyntaxException('%s is not integer anyway' % value)
        else:
            return None


def string_from_string(value, is_strict=False):
    if not isinstance(value, six.string_types):
        if is_strict:
            raise SphinxQLSyntaxException('%s is not string anyway' % value)
        else:
            return None
    return value


def list_of_integers_only(sequence, is_strict=False):
    cleaned_sequence = []
    for orig_value in sequence:
        clean_value = int_from_digit(orig_value, is_strict)
        if clean_value is not None:
            cleaned_sequence.append(clean_value)
    return cleaned_sequence


def list_of_strings_only(sequence, is_strict=False):
    cleaned_sequence = []
    for orig_value in sequence:
        clean_value = string_from_string(orig_value, is_strict)
        if clean_value is not None:
            cleaned_sequence.append(clean_value)
    return cleaned_sequence


def sparse_free_sequence(sequence):
    return [
        x for x in sequence
        if (
            isinstance(x, six.string_types) and bool(x.strip())
            or not isinstance(x, six.string_types) and bool(x)
        )
    ]


def unix_timestamp(datetime):
    return str(int(time.mktime(datetime.timetuple())))


class BaseSearchConfig(object):
    DEBUG = True
    WITH_META = True
    WITH_STATUS = True
    POOL_SIZE = 5
    SEARCHD_CONNECTION = {
        'host': '127.0.0.1',
        'port': 9306,
    }
    # For future usage
    SNIPPETS_DEFAULTS = {
        'before_match': '<strong>',
        'after_match': '</strong>',
    }



"""
    sphinxit.core.processor
    ~~~~~~~~~~~~~~~~~~~~~~~

    Implements SphinxQL expression processing.

    :copyright: (c) 2013 by Roman Semirook.
    :license: BSD, see LICENSE for more details.
"""

from __future__ import unicode_literals

try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict

from copy import deepcopy

import six

from sphinxit.core.helpers import sparse_free_sequence
from sphinxit.core.nodes import (
    SelectFromContainer,
    AggregateObject,
    UpdateSetNode,
    FiltersContainer,
    LimitNode,
    GroupByNode,
    WithinGroupOrderByNode,
    OrderByContainer,
    OptionsContainer,
    OR,
    SnippetsOptionsContainer,
    SnippetsQueryNode,
    RawAttr
)
from sphinxit.core.mixins import ConfigMixin
from sphinxit.core.constants import NODES_ORDER
from sphinxit.core.connector import SphinxConnector


class LazySelectTree(ConfigMixin):

    def __init__(self, indexes):
        self._indexes = indexes
        self._nodes = OrderedDict([
            ('SelectFrom', None),
            ('UpdateSet', None),
            ('Where', None),
            ('GroupBy', None),
            ('OrderBy', None),
            ('WithinGroupOrderBy', None),
            ('Limit', None),
            ('Options', None),
        ])
        super(LazySelectTree, self).__init__()

    def __bool__(self):
        return bool(self._nodes['SelectFrom'] or self._nodes['UpdateSet'])

    def copy(self):
        new_tree = LazySelectTree(self._indexes).with_config(self.config)
        new_tree._indexes = self._indexes[:]
        new_tree._nodes = deepcopy(self._nodes)
        return new_tree

    @property
    def SelectFrom(self):
        if self._nodes['SelectFrom'] is None:
            select_container = SelectFromContainer(indexes=self._indexes).with_config(self.config)
            self._nodes['SelectFrom'] = select_container
        return self._nodes['SelectFrom']

    @property
    def Where(self):
        if self._nodes['Where'] is None:
            filters_container = FiltersContainer().with_config(self.config)
            self._nodes['Where'] = filters_container
        return self._nodes['Where']

    @property
    def GroupBy(self):
        if self._nodes['GroupBy'] is None:
            group_by_node = GroupByNode().with_config(self.config)
            self._nodes['GroupBy'] = group_by_node
        return self._nodes['GroupBy']

    @property
    def OrderBy(self):
        if self._nodes['OrderBy'] is None:
            order_by_container = OrderByContainer().with_config(self.config)
            self._nodes['OrderBy'] = order_by_container
        return self._nodes['OrderBy']

    @property
    def WithinGroupOrderBy(self):
        if self._nodes['WithinGroupOrderBy'] is None:
            within_group_order_by_node = WithinGroupOrderByNode().with_config(self.config)
            self._nodes['WithinGroupOrderBy'] = within_group_order_by_node
        return self._nodes['WithinGroupOrderBy']

    @property
    def Limit(self):
        if self._nodes['Limit'] is None:
            self._nodes['Limit'] = LimitNode().with_config(self.config)
        return self._nodes['Limit']

    @property
    def Options(self):
        if self._nodes['Options'] is None:
            options_container = OptionsContainer().with_config(self.config)
            self._nodes['Options'] = options_container
        return self._nodes['Options']

    @property
    def UpdateSet(self):
        if self._nodes['UpdateSet'] is None:
            update_set_node = UpdateSetNode(indexes=self._indexes).with_config(self.config)
            self._nodes['UpdateSet'] = update_set_node
        return self._nodes['UpdateSet']

    def is_update(self):
        return self._nodes['UpdateSet'] is not None

    def get_select_nodes(self):
        if self._nodes['SelectFrom'] is None:
            select_container = SelectFromContainer(indexes=self._indexes).with_config(self.config)
            self._nodes['SelectFrom'] = select_container

        return [self._nodes[n] for n in NODES_ORDER.select]

    def get_update_nodes(self):
        return [self._nodes[n] for n in NODES_ORDER.update]


class LazySnippetsTree(ConfigMixin):
    _template = 'CALL SNIPPETS ({conditions})'

    def __init__(self, index):
        self._index = index
        self._snippets_syntax = OrderedDict([
            ('SnippetQuery', None),
            ('Options', None),
        ])
        super(LazySnippetsTree, self).__init__()

    def __bool__(self):
        return bool(self._snippets_syntax['SnippetQuery'])

    @property
    def SnippetQuery(self):
        if self._snippets_syntax['SnippetQuery'] is None:
            self._snippets_syntax['SnippetQuery'] = (
                SnippetsQueryNode(index=self._index).with_config(self.config)
            )
        return self._snippets_syntax['SnippetQuery']

    @property
    def Options(self):
        if self._snippets_syntax['Options'] is None:
            self._snippets_syntax['Options'] = (
                SnippetsOptionsContainer().with_config(self.config)
            )
        return self._snippets_syntax['Options']

    def lex(self):
        return self._template.format(
            conditions=', '.join([
                x.lex()
                for x in sparse_free_sequence(self._snippets_syntax.values())
            ])
        )


def copy_tree(method):
    def wrapper(self, *args, **kwargs):
        self_copy = self.__class__(self.indexes, self.config, self.connector)
        self_copy._nodes = self._nodes.copy()
        return method(self_copy, *args, **kwargs)
    return wrapper


class Search(ConfigMixin):

    def __init__(self, indexes, config, connector=None):
        super(Search, self).__init__()
        self._nodes = LazySelectTree(indexes=indexes).with_config(config)
        self.indexes = indexes  # TODO: add check it's iterable
        self.config = config
        self.connector = connector or SphinxConnector(config)

    @copy_tree
    def select(self, *args, **kwargs):
        if args:
            for field in args:
                if isinstance(field, six.string_types):
                    self._nodes.SelectFrom.add_field(field)
                if isinstance(field, (tuple, list)) and len(field) == 2:
                    field, alias = field
                    self._nodes.SelectFrom.add_alias(field, alias)
                if isinstance(field, AggregateObject):
                    self._nodes.SelectFrom.add_aggregation(field)
                if isinstance(field, RawAttr):
                    self._nodes.SelectFrom.add_raw_attr(field)
        if kwargs:
            for field, alias in kwargs.items():
                self._nodes.SelectFrom.add_alias(field, alias)

        return self

    @copy_tree
    def update(self, **kwargs):
        if kwargs:
            for field, value in kwargs.items():
                self._nodes.UpdateSet.update(field, value)
        return self

    @copy_tree
    def match(self, query, raw=False):
        if not raw:
            self._nodes.Where.add_query(query)
        else:
            self._nodes.Where.add_raw_query(query)

        return self

    @copy_tree
    def filter(self, *args, **kwargs):
        if args:
            for cond in args:
                if isinstance(cond, OR):
                    self._nodes.SelectFrom.add_or(cond)
                    self._nodes.Where.add_condition('cnd__gt', 0)
        if kwargs:
            self._nodes.Where.add_conditions(**kwargs)

        return self

    @copy_tree
    def limit(self, offset=None, limit=None):
        self._nodes.Limit.set_range(offset, limit)
        return self

    @copy_tree
    def group_by(self, field):
        self._nodes.GroupBy.by_field(field)
        return self

    @copy_tree
    def within_group_order_by(self, field, ordering=None):
        self._nodes.WithinGroupOrderBy.by_field(field, ordering)
        return self

    @copy_tree
    def order_by(self, field, ordering=None):
        self._nodes.OrderBy.by_field(field, ordering)
        return self

    @copy_tree
    def options(self, **kwargs):
        self._nodes.Options.set_options(**kwargs)
        return self

    @copy_tree
    def named(self, name):
        self._name = name
        return self

    def lex(self):
        if self._nodes.is_update():
            actual_nodes = self._nodes.get_update_nodes()
        else:
            actual_nodes = self._nodes.get_select_nodes()

        return ' '.join([
            x.lex() for x in sparse_free_sequence(actual_nodes)
        ])

    def ask(self, subqueries=None):
        query_batch = [(self.lex(), getattr(self, '_name', 'result'))]
        if subqueries is not None:
            query_batch.extend([
                (s_inst.lex(), getattr(s_inst, '_name', 'result_%s' % id(s_inst)))
                for s_inst in subqueries
            ])
        return self.connector.execute(query_batch)


class Snippet(ConfigMixin):

    def __init__(self, index=None, config=None, connector=None):
        super(Snippet, self).__init__()
        self._snippets_tree = LazySnippetsTree(index=index).with_config(config)
        self.index = index
        self.config = config
        self.connector = connector or SphinxConnector(config)

    def from_data(self, *args):
        self._snippets_tree.SnippetQuery.add_data(*args)
        return self

    def for_query(self, query):
        self._snippets_tree.SnippetQuery.add_query(query)
        return self

    def options(self, **kwargs):
        self._snippets_tree.Options.set_options(**kwargs)
        return self

    def lex(self):
        return self._snippets_tree.lex()

    def ask(self):
        return self.connector.execute(self.lex())
